"""Scryfall object models."""

import datetime as dt
from decimal import Decimal
from enum import Enum
from typing import Generic, TypeAlias, TypeVar
from uuid import UUID

from msgspec import Struct


class ScryColor(str, Enum):
    """Enum for https://scryfall.com/docs/api/colors#color-arrays."""

    WHITE = "W"
    BLUE = "U"
    BLACK = "B"
    RED = "R"
    GREEN = "G"
    COLORLESS = "C"
    TAP = "T"


class ScrySetType(str, Enum):
    """Enum for https://scryfall.com/docs/api/sets#set-types."""

    CORE = "core"
    EXPANSION = "expansion"
    ETERNAL = "eternal"
    MASTERS = "masters"
    MASTERPIECE = "masterpiece"
    FROM_THE_VAULT = "from_the_vault"
    SPELLBOOK = "spellbook"
    PREMIUM_DECK = "premium_deck"
    DUEL_DECK = "duel_deck"
    DRAFT_INNOVATION = "draft_innovation"
    TREASURE_CHEST = "treasure_chest"
    COMMANDER = "commander"
    PLANECHASE = "planechase"
    ARCHENEMY = "archenemy"
    VANGUARD = "vanguard"
    FUNNY = "funny"
    STARTER = "starter"
    BOX = "box"
    PROMO = "promo"
    TOKEN = "token"  # noqa: S105
    MEMORABILIA = "memorabilia"
    ALCHEMY = "alchemy"
    ARSENAL = "arsenal"
    MINIGAME = "minigame"


class ScryCardLayout(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#layout."""

    ADVENTURE = "adventure"
    ART_SERIES = "art_series"
    AUGMENT = "augment"
    CASE = "case"
    CLASS = "class"
    DOUBLE_FACED_TOKEN = "double_faced_token"  # noqa: S105
    DOUBLE_SIDED = "double_sided"
    EMBLEM = "emblem"
    FLIP = "flip"
    HOST = "host"
    LEVELER = "leveler"
    MELD = "meld"
    MODAL_DFC = "modal_dfc"
    MUTATE = "mutate"
    NORMAL = "normal"
    PLANAR = "planar"
    PROTOTYPE = "prototype"
    REVERSIBLE_CARD = "reversible_card"
    SAGA = "saga"
    SCHEME = "scheme"
    SPLIT = "split"
    TOKEN = "token"  # noqa: S105
    TRANSFORM = "transform"
    VANGUARD = "vanguard"


class ScryCardFrame(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#frames."""

    Y1993 = "1993"
    Y1997 = "1997"
    Y2003 = "2003"
    Y2015 = "2015"
    FUTURE = "future"


class ScryFrameEffect(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#frame-effects."""

    NONE = ""

    BORDERLESS = "borderless"
    COLORSHIFTED = "colorshifted"
    COMPANION = "companion"
    COMPASSLANDDFC = "compasslanddfc"
    CONVERTDFC = "convertdfc"
    DEVOID = "devoid"
    DRAFT = "draft"
    ENCHANTMENT = "enchantment"
    ETCHED = "etched"
    EXTENDEDART = "extendedart"
    FANDFC = "fandfc"
    FULLART = "fullart"
    GILDED = "gilded"
    GRAVESTONE = "gravestone"
    INVERTED = "inverted"
    LEGENDARY = "legendary"
    LESSON = "lesson"
    MIRACLE = "miracle"
    MOONELDRAZIDFC = "mooneldrazidfc"
    MOONREVERSEMOONDFC = "moonreversemoondfc"
    NYXBORN = "nyxborn"
    NYXTOUCHED = "nyxtouched"
    ORIGINPWDFC = "originpwdfc"
    PROMO = "promo"
    SHATTEREDGLASS = "shatteredglass"
    SHOWCASE = "showcase"
    SNOW = "snow"
    SPREE = "spree"
    STAMPED = "stamped"
    SUNMOONDFC = "sunmoondfc"
    TEXTLESS = "textless"
    THICK = "thick"
    TOMBSTONE = "tombstone"
    UPSIDEDOWNDFC = "upsidedowndfc"
    VEHICLE = "vehicle"
    WAXINGANDWANINGMOONDFC = "waxingandwaningmoondfc"


class ScryBorderColor(str, Enum):
    """Enum for card border_color."""

    BLACK = "black"
    BORDERLESS = "borderless"
    GOLD = "gold"
    SILVER = "silver"
    WHITE = "white"
    YELLOW = "yellow"


class ScryFinish(str, Enum):
    """Enum for card finishes."""

    FOIL = "foil"
    NONFOIL = "nonfoil"
    ETCHED = "etched"
    GLOSSY = "glossy"


class ScryImageStatus(str, Enum):
    """Enum for card image_status."""

    MISSING = "missing"
    PLACEHOLDER = "placeholder"
    LOWRES = "lowres"
    HIGHRES_SCAN = "highres_scan"


class ScryGame(str, Enum):
    """Enum for card games."""

    PAPER = "paper"
    ARENA = "arena"
    MTGO = "mtgo"
    SEGA = "sega"
    ASTRAL = "astral"


class ScryRarity(str, Enum):
    """Enum for card rarity."""

    COMMON = "common"
    UNCOMMON = "uncommon"
    RARE = "rare"
    MYTHIC = "mythic"
    SPECIAL = "special"
    BONUS = "bonus"


class ScrySecurityStamp(str, Enum):
    """Enum for card security_stamp."""

    OVAL = "oval"
    TRIANGLE = "triangle"
    ACORN = "acorn"
    ARENA = "arena"
    CIRCLE = "circle"
    HEART = "heart"


class ScryFormat(str, Enum):
    """Enum for card legalities keys."""

    ALCHEMY = "alchemy"
    BRAWL = "brawl"
    COMMANDER = "commander"
    DUEL = "duel"
    EXPLORER = "explorer"
    FRONTIER = "frontier"
    FUTURE = "future"
    GLADIATOR = "gladiator"
    HISTORIC = "historic"
    HISTORICBRAWL = "historicbrawl"
    LEGACY = "legacy"
    MODERN = "modern"
    OATHBREAKER = "oathbreaker"
    OLDSCHOOL = "oldschool"
    PAUPER = "pauper"
    PAUPERCOMMANDER = "paupercommander"
    PENNY = "penny"
    PIONEER = "pioneer"
    PREDH = "predh"
    PREMODERN = "premodern"
    STANDARD = "standard"
    STANDARDBRAWL = "standardbrawl"
    TIMELESS = "timeless"
    VINTAGE = "vintage"


class ScryLegality(str, Enum):
    """Enum for card legalities values."""

    LEGAL = "legal"
    NOT_LEGAL = "not_legal"
    RESTRICTED = "restricted"
    BANNED = "banned"


class ScryMigrationStrategy(str, Enum):
    """Enum for migration strategy values."""

    MERGE = "merge"
    DELETE = "delete"


class ScrySet(
    Struct,
    tag_field="object",
    tag="set",
    kw_only=True,
    omit_defaults=True,
):
    """Model for https://scryfall.com/docs/api/sets."""

    id: UUID
    code: str
    mtgo_code: str | None = None
    arena_code: str | None = None
    tcgplayer_id: int | None = None
    name: str
    set_type: ScrySetType
    released_at: dt.date | None = None
    block_code: str | None = None
    block: str | None = None
    parent_set_code: str | None = None
    card_count: int
    printed_size: int | None = None
    digital: bool
    foil_only: bool
    nonfoil_only: bool | None = None
    icon_svg_uri: str
    search_uri: str
    scryfall_uri: str
    uri: str


class ScryRelatedCard(
    Struct,
    tag_field="object",
    tag="related_card",
    kw_only=True,
    omit_defaults=True,
):
    """Model for https://scryfall.com/docs/api/cards#related-card-objects."""

    id: UUID
    component: str
    name: str
    type_line: str
    uri: str


class ScryCardFace(
    Struct,
    tag_field="object",
    tag="card_face",
    kw_only=True,
    omit_defaults=True,
):
    """Model for https://scryfall.com/docs/api/cards#card-face-objects."""

    artist: str | None = None
    artist_id: UUID | None = None
    cmc: float | None = None
    color_indicator: list[ScryColor] | None = None
    colors: list[ScryColor] | None = None
    flavor_name: str | None = None
    flavor_text: str | None = None
    illustration_id: UUID | None = None
    image_uris: dict[str, str] | None = None
    layout: ScryCardLayout | None = None
    loyalty: str | None = None
    mana_cost: str
    name: str
    oracle_id: UUID | None = None
    oracle_text: str | None = None
    power: str | None = None
    printed_name: str | None = None
    printed_text: str | None = None
    printed_type_line: str | None = None
    toughness: str | None = None
    type_line: str | None = None
    watermark: str | None = None


class CardPreviewBlock(Struct):
    """Model for card preview block."""

    source: str
    source_uri: str
    previewed_at: dt.date


class ScryCard(
    Struct,
    tag_field="object",
    tag="card",
    kw_only=True,
    omit_defaults=True,
):
    """Model for https://scryfall.com/docs/api/cards."""

    # Core Card Fields
    arena_id: int | None = None
    id: UUID
    lang: str
    mtgo_id: int | None = None
    mtgo_foil_id: int | None = None
    multiverse_ids: list[int] | None = None
    tcgplayer_id: int | None = None
    tcgplayer_etched_id: int | None = None
    cardmarket_id: int | None = None
    oracle_id: UUID | None = None
    prints_search_uri: str
    rulings_uri: str
    scryfall_uri: str
    uri: str
    # Gameplay Fields
    all_parts: list[ScryRelatedCard] | None = None
    card_faces: list[ScryCardFace] | None = None
    cmc: float | None = None
    colors: list[ScryColor] | None = None
    color_identity: list[ScryColor]
    color_indicator: list[ScryColor] | None = None
    edhrec_rank: int | None = None
    foil: bool
    hand_modifier: str | None = None
    keywords: list[str]
    layout: ScryCardLayout
    legalities: dict[ScryFormat, ScryLegality]
    life_modifier: str | None = None
    loyalty: str | None = None
    mana_cost: str | None = None
    name: str
    nonfoil: bool
    oracle_text: str | None = None
    oversized: bool
    penny_rank: int | None = None
    power: str | None = None
    produced_mana: list[str] | None = None
    reserved: bool
    toughness: str | None = None
    type_line: str | None = None
    # Print Fields
    artist: str | None = None
    artist_ids: list[UUID] | None = None
    booster: bool
    border_color: ScryBorderColor
    card_back_id: UUID | None = None
    collector_number: str
    content_warning: bool | None = None
    digital: bool
    finishes: list[ScryFinish]
    flavor_name: str | None = None
    flavor_text: str | None = None
    frame_effect: ScryFrameEffect | None = None
    frame_effects: list[ScryFrameEffect] | None = None
    frame: ScryCardFrame
    full_art: bool
    games: list[ScryGame]
    highres_image: bool
    illustration_id: UUID | None = None
    image_status: ScryImageStatus
    image_uris: dict[str, str] | None = None
    prices: dict[str, Decimal | None] | None  # TODO: enum keys=None
    printed_name: str | None = None
    printed_text: str | None = None
    printed_type_line: str | None = None
    promo: bool
    promo_types: list[str] | None = None
    purchase_uris: dict[str, str] | None = None
    rarity: ScryRarity
    related_uris: dict[str, str] | None = None
    released_at: dt.date
    reprint: bool
    scryfall_set_uri: str
    set_name: str
    set_search_uri: str
    set_type: str
    set_uri: str
    set: str
    set_id: UUID
    story_spotlight: bool
    textless: bool
    variation: bool
    variation_of: UUID | None = None
    security_stamp: ScrySecurityStamp | None = None
    watermark: str | None = None
    preview: CardPreviewBlock | None = None


class ScryBulkData(
    Struct,
    tag_field="object",
    tag="bulk_data",
    kw_only=True,
    omit_defaults=True,
):
    """Model for https://scryfall.com/docs/api/bulk-data."""

    id: UUID
    uri: str
    type: str
    name: str
    description: str
    download_uri: str
    updated_at: dt.datetime
    compressed_size: int | None = None
    content_type: str
    content_encoding: str


class ScryMigration(
    Struct,
    tag_field="object",
    tag="migration",
    kw_only=True,
    omit_defaults=True,
):
    """Model for https://scryfall.com/docs/api/migrations."""

    id: UUID
    uri: str
    performed_at: dt.date
    migration_strategy: ScryMigrationStrategy
    old_scryfall_id: UUID
    new_scryfall_id: UUID | None = None
    note: str | None = None


ScryListable: TypeAlias = ScryBulkData | ScryCard | ScryMigration | ScrySet

_ScryListableT = TypeVar("_ScryListableT", bound=ScryListable)


class ScryList(
    Struct,
    Generic[_ScryListableT],
    tag_field="object",
    tag="list",
    kw_only=True,
    omit_defaults=True,
):
    """Model for https://scryfall.com/docs/api/lists."""

    data: list[_ScryListableT]
    has_more: bool
    next_page: str | None = None
    total_cards: int | None = None
    warnings: list[str] | None = None



"""XLSX serializer."""

import collections
import datetime as dt
import importlib.util
import itertools
import string
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any, ClassVar

import openpyxl
from openpyxl.styles.numbers import FORMAT_CURRENCY_USD_SIMPLE
from openpyxl.workbook.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet

from mtg_ssm.containers import counts
from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.indexes import Oracle, ScryfallDataIndex
from mtg_ssm.mtg import util
from mtg_ssm.scryfall.models import ScryCard, ScrySet
from mtg_ssm.serialization import interface

HAS_LXML = importlib.util.find_spec("lxml") is not None

ALL_SETS_SHEET_HEADER: Sequence[str] = [
    "code",
    "name",
    "release",
    "block",
    "type",
    "printed",
    "have",
    "unique",
    "playsets",
    "extras",
    "nonfoil",
    "foil",
    "value",
]

ALL_SETS_SHEET_TOTALS: Sequence[str | None] = ["Total", None, None, None, None] + [
    f"=SUM({c}3:{c}65535)" for c in "FGHIJKLM"
]


def _card_set_sort_key(cset: ScrySet) -> tuple[dt.date, str, int, str]:
    released_at = cset.released_at or dt.date.min
    return (
        released_at,
        cset.parent_set_code or cset.code,
        0 if cset.parent_set_code is None else 1,
        cset.code,
    )


def create_all_sets(sheet: Worksheet, index: ScryfallDataIndex) -> None:
    """Create all sets sheet from card_db."""
    sheet.title = "All Sets"
    sheet.append(ALL_SETS_SHEET_HEADER)
    sheet.append(ALL_SETS_SHEET_TOTALS)
    for card_set in sorted(index.setcode_to_set.values(), key=_card_set_sort_key):
        setcode = card_set.code.upper()
        card_count = len(index.setcode_to_cards[card_set.code])
        have_range = _setsheet_data_range(setcode, "have", card_count)
        nonfoil_range = _setsheet_data_range(setcode, "nonfoil", card_count)
        foil_range = _setsheet_data_range(setcode, "foil", card_count)
        value_range = _setsheet_data_range(setcode, "value", card_count)
        row = [
            setcode,
            card_set.name,
            card_set.released_at,
            card_set.block,
            card_set.set_type.value,
            card_count,
            f"=SUM({have_range})",
            f'=COUNTIF({have_range},">0")',
            f'=COUNTIF({have_range},">=4")',
            f'=SUMIF({have_range},">4")-4*COUNTIF({have_range},">4")',
            f"=SUM({nonfoil_range})",
            f"=SUM({foil_range})",
            f"=SUM({value_range})",
        ]
        sheet.append(row)


def style_all_sets(sheet: Worksheet) -> None:
    """Apply styles to the all sets sheet."""
    sheet.freeze_panes = "C3"
    col_width_hidden_format = [
        ("A", 8, False, None),
        ("B", 30, False, None),
        ("C", 12, True, None),
        ("D", 22, True, None),
        ("E", 15, True, None),
        ("F", 7, False, None),
        ("G", 6, False, None),
        ("H", 7, False, None),
        ("I", 8, False, None),
        ("J", 6, False, None),
        ("K", 7, False, None),
        ("L", 5, False, None),
        ("M", 10, False, FORMAT_CURRENCY_USD_SIMPLE),
    ]
    for col, width, hidden, number_format in col_width_hidden_format:
        cdim = sheet.column_dimensions[col]
        cdim.width = width
        cdim.hidden = hidden
        if number_format is not None:
            cdim.number_format = number_format


def create_haverefs(index: ScryfallDataIndex, setcode: str, cards: Sequence[ScryCard]) -> str:
    """Create a reference to or sum of the have cell(s) for printings in a single set."""
    setcode = setcode.upper()
    rownums = sorted(index.id_to_setindex[card.id] + ROW_OFFSET for card in cards)
    haverefs = []
    # cluster contiguous rownums
    for _, group in itertools.groupby(enumerate(rownums), lambda x: x[0] - x[1]):
        cluster = [g[1] for g in group]
        cells = [_setsheet_col("have") + str(r) for r in cluster]
        if cells[0] == cells[-1]:
            haverefs.append(f"'{setcode}'!{cells[0]}")
        else:
            haverefs.append(f"'{setcode}'!{cells[0]}:{cells[-1]}")
    if len(haverefs) == 1 and ":" not in haverefs[0]:
        return haverefs[0]
    return f"SUM({','.join(haverefs)})"


def get_references(
    index: ScryfallDataIndex, card_name: str, exclude_sets: set[str] | None = None
) -> str | None:
    """Get an equation for the references to a card."""
    if util.is_strict_basic(card_name):
        return None  # Basics are so prolific that they overwhelm Excel

    if exclude_sets is None:
        exclude_sets = set()

    set_to_cards: dict[str, list[ScryCard]] = collections.defaultdict(list)
    for other_card in index.name_to_cards[card_name]:
        if other_card.set not in exclude_sets:
            set_to_cards[other_card.set].append(other_card)

    set_to_haveref = {k: create_haverefs(index, k, v) for k, v in set_to_cards.items()}
    if not set_to_haveref:
        return None

    references = []
    for card_set in sorted(
        set_to_haveref,
        key=lambda setcode: _card_set_sort_key(index.setcode_to_set[setcode]),
    ):
        reference = (
            f'IF({set_to_haveref[card_set]}>0,"{card_set.upper()}:"&{set_to_haveref[card_set]},"")'
        )
        references.append(reference)
    if not references:
        return None
    if len(references) == 1:
        return "=" + references[0]
    return f'=_xlfn.TEXTJOIN(", ",1,{",".join(references)})'


ALL_CARDS_SHEET_HEADER = ["name", "have"]  # TODO: add list of sets


def create_all_cards(sheet: Worksheet, index: ScryfallDataIndex) -> None:
    """Create all cards sheet from card_db."""
    sheet.title = "All Cards"
    sheet.append(ALL_CARDS_SHEET_HEADER)
    for name in sorted(index.name_to_cards):
        row = [name, get_references(index, name)]
        sheet.append(row)


def style_all_cards(sheet: Worksheet) -> None:
    """Apply styles to the all cards sheet."""
    sheet.freeze_panes = "A2"
    col_width_hidden = [
        ("A", 28, False),
        ("B", 48, False),
    ]
    for col, width, hidden in col_width_hidden:
        cdim = sheet.column_dimensions[col]
        cdim.width = width
        cdim.hidden = hidden


SET_SHEET_HEADER = (
    [
        "have",
        "value",
        "name",
        "number",
        "scryfall_id",
        "artist",
        "price",
        "foil_price",
    ]
    + [ct.value for ct in counts.CountType]
    + ["others"]
)


def _setsheet_col(column_header: str) -> str:
    return string.ascii_uppercase[SET_SHEET_HEADER.index(column_header)]


def _setsheet_data_range(setcode: str, column_header: str, row_count: int) -> str:
    """Create an absolute worksheet range spanning the populated card rows for a column."""
    col = _setsheet_col(column_header)
    first_row = ROW_OFFSET
    last_row = ROW_OFFSET + row_count - 1
    return f"'{setcode}'!{col}{first_row}:{col}{last_row}"


COUNT_COLS = [_setsheet_col(ct) for ct in counts.CountType]
HAVE_TMPL = "=" + "+".join(col + "{rownum}" for col in COUNT_COLS)
HAVE_TMPL = (
    "="
    + _setsheet_col(counts.CountType.NONFOIL)
    + "{rownum}"
    + "+"
    + _setsheet_col(counts.CountType.FOIL)
    + "{rownum}"
)
VALUE_TMPL = (
    "="
    + _setsheet_col(counts.CountType.NONFOIL)
    + "{rownum}"
    + "*"
    + _setsheet_col("price")
    + "{rownum}"
    + "+"
    + _setsheet_col(counts.CountType.FOIL)
    + "{rownum}"
    + "*"
    + _setsheet_col("foil_price")
    + "{rownum}"
)
ROW_OFFSET = 2


def create_set_sheet(sheet: Worksheet, collection: MagicCollection, setcode: str) -> None:
    """Populate sheet with card information from a given set."""
    index = collection.oracle.index

    sheet.append(SET_SHEET_HEADER)
    sheet.title = setcode.upper()

    for card in index.setcode_to_cards[setcode]:
        rownum = ROW_OFFSET + index.id_to_setindex[card.id]
        row: list[Any | None] = [
            HAVE_TMPL.format(rownum=rownum),
            VALUE_TMPL.format(rownum=rownum),
            card.name,
            card.collector_number,
            str(card.id),
            card.artist,
            (card.prices or {}).get("usd", None),
            (card.prices or {}).get("usd_foil", None),
        ]
        card_counts = collection.counts.get(card.id, {})
        for count_type in counts.CountType:
            row.append(card_counts.get(count_type))
        row.append(get_references(index, card.name, exclude_sets={setcode}))
        sheet.append(row)


def style_set_sheet(sheet: Worksheet) -> None:
    """Apply styles to a set sheet."""
    sheet.freeze_panes = "E2"
    col_width_hidden_format = [
        ("A", 5, False, None),
        ("B", 9, False, FORMAT_CURRENCY_USD_SIMPLE),
        ("C", 24, False, None),
        ("D", 7, False, None),
        ("E", 10, True, None),
        ("F", 20, True, None),
        ("G", 9, True, FORMAT_CURRENCY_USD_SIMPLE),
        ("H", 9, True, FORMAT_CURRENCY_USD_SIMPLE),
        ("I", 7, False, None),
        ("J", 6, False, None),
        ("K", 10, False, None),
    ]
    for col, width, hidden, number_format in col_width_hidden_format:
        cdim = sheet.column_dimensions[col]
        cdim.width = width
        cdim.hidden = hidden
        if number_format is not None:
            cdim.number_format = number_format


def rows_from_sheet(sheet: Worksheet) -> Iterable[dict[str, str]]:
    """Read rows from an xlsx worksheet as dicts."""
    header_row, *rows = iter(sheet.rows)
    header = [cell.value for cell in header_row]
    for row in rows:
        values = [cell.value for cell in row]
        if any(v is not None for v in values):
            yield dict(zip(header, values, strict=False), set=str(sheet.title))


def rows_for_workbook(book: Workbook, *, skip_sheets: set[str] | None) -> Iterable[dict[str, str]]:
    """Read rows from an xlsx workbook as dicts."""
    if skip_sheets is None:
        skip_sheets = set()
    for sheet in book.worksheets:
        if sheet.title in skip_sheets:
            continue
        yield from rows_from_sheet(sheet)


class XlsxDialect(interface.SerializationDialect):
    """excel xlsx collection."""

    extension: ClassVar[str] = "xlsx"
    dialect: ClassVar[str] = "xlsx"

    def write(self, path: Path, collection: MagicCollection) -> None:
        """Write collection to an xlsx file."""
        workbook = openpyxl.Workbook(write_only=HAS_LXML)

        all_sets_sheet = workbook.create_sheet()
        style_all_sets(all_sets_sheet)
        create_all_sets(all_sets_sheet, collection.oracle.index)

        all_cards_sheet = workbook.create_sheet()
        style_all_cards(all_cards_sheet)
        create_all_cards(all_cards_sheet, collection.oracle.index)

        setcodes = [
            s.code
            for s in sorted(
                collection.oracle.index.setcode_to_set.values(),
                key=_card_set_sort_key,
            )
        ]

        for setcode in setcodes:
            set_sheet = workbook.create_sheet()
            style_set_sheet(set_sheet)
            create_set_sheet(set_sheet, collection, setcode)

        if not HAS_LXML:
            # write_only mode does no create a default sheet but read/write mode does
            # if we do not have lxml, we are in read/write mode and must delete the
            # default sheet
            del workbook["Sheet"]
        workbook.save(str(path))

    def read(self, path: Path, oracle: Oracle) -> MagicCollection:
        """Read collection from an xlsx file."""
        workbook = openpyxl.load_workbook(filename=str(path), read_only=True)
        reader = rows_for_workbook(workbook, skip_sheets={"All Sets", "All Cards"})
        card_counts = counts.aggregate_card_counts(reader, oracle)
        return MagicCollection(oracle=oracle, counts=card_counts)



"""Interface definition for serializers."""

import abc
from pathlib import Path
from typing import ClassVar

from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.indexes import Oracle


class Error(Exception):
    """Base error for serializers."""


class UnknownDialectError(Exception):
    """Raised when an (extension, dialect) pair is requested."""


class DeserializationError(Error):
    """Raised when there is an error reading counts from a file."""


class SerializationDialect(metaclass=abc.ABCMeta):
    """Abstract interface for mtg ssm serialization dialect."""

    _EXT_DIALECT_DOC: ClassVar[set[tuple[str, str, str]]] = set()
    _EXT_DIALECT_TO_IMPL: ClassVar[dict[tuple[str, str], type["SerializationDialect"]]] = {}

    extension: ClassVar[str | None] = None
    dialect: ClassVar[str | None] = None

    def __init_subclass__(cls: type["SerializationDialect"]) -> None:
        super().__init_subclass__()
        if cls.extension is not None and cls.dialect is not None:
            cls._EXT_DIALECT_DOC.add((cls.extension, cls.dialect, cls.__doc__ or cls.__name__))
            cls._EXT_DIALECT_TO_IMPL[(cls.extension, cls.dialect)] = cls

    @abc.abstractmethod
    def write(self, path: Path, collection: MagicCollection) -> None:
        """Write print counts to a file."""

    @abc.abstractmethod
    def read(self, path: Path, oracle: Oracle) -> MagicCollection:
        """Read print counts from file."""

    @classmethod
    def dialects(
        cls: type["SerializationDialect"],
    ) -> list[tuple[str, str | None, str | None]]:
        """List of (extension, dialect, description) of registered dialects."""
        return sorted((ext, dial or "", doc or "") for ext, dial, doc in cls._EXT_DIALECT_DOC)

    @classmethod
    def by_extension(
        cls: type["SerializationDialect"],
        extension: str,
        dialect_mappings: dict[str, str],
    ) -> type["SerializationDialect"]:
        """Get a serializer class for a given extension and dialect mapping."""
        dialect = dialect_mappings.get(extension, extension)
        try:
            return cls._EXT_DIALECT_TO_IMPL[(extension, dialect)]
        except KeyError as err:
            msg = f'File extension: "{extension}" dialect: "{dialect}" not found in registry'
            raise UnknownDialectError(msg) from err



"""Tests for mtg_ssm.scryfall.models."""

import msgspec

from mtg_ssm.scryfall.models import ScrySet, ScrySetType


def test_scryset_supports_eternal_set_type() -> None:
    payload = b"""{
        "object": "set",
        "id": "f0d89b4b-1f4f-4b22-a201-e1e84b4a5d5f",
        "code": "tmc",
        "name": "Teenage Mutant Ninja Turtles Eternal",
        "set_type": "eternal",
        "card_count": 8,
        "digital": false,
        "foil_only": false,
        "icon_svg_uri": "https://svgs.scryfall.io/sets/default.svg?1694404800",
        "search_uri": "https://api.scryfall.com/cards/search?order=set&q=e%3Atmc&unique=prints",
        "scryfall_uri": "https://scryfall.com/sets/tmc",
        "uri": "https://api.scryfall.com/sets/tmc"
    }"""

    scryset = msgspec.json.decode(payload, type=ScrySet)

    assert scryset.set_type is ScrySetType.ETERNAL



"""Tests for mtg_ssm.containers.bundles."""

from collections import Counter
from uuid import UUID

from mtg_ssm.containers import bundles
from mtg_ssm.scryfall.models import ScryCardLayout, ScrySetType


def test_remove_digital(scryfall_data: bundles.ScryfallDataSet) -> None:
    set_codes = {s.code for s in scryfall_data.sets}
    card_ids = {c.id for c in scryfall_data.cards}
    card_names = {c.name for c in scryfall_data.cards}

    assert "vma" in set_codes
    assert UUID("116ec16c-3b4b-45be-83c8-333bccc29e35") in card_ids

    assert "khm" in set_codes
    assert "Cosmos Elixir" in card_names
    assert "A-Cosmos Elixir" in card_names

    assert "yecl" in set_codes
    assert "Thoughtsteal Sprites" in card_names

    digital_removed = bundles.filter_cards_and_sets(scryfall_data, exclude_digital=True)
    set_codes2 = {s.code for s in digital_removed.sets}
    card_ids2 = {c.id for c in digital_removed.cards}
    card_names2 = {c.name for c in digital_removed.cards}

    assert "vma" not in set_codes2
    assert UUID("116ec16c-3b4b-45be-83c8-333bccc29e35") not in card_ids2

    assert "khm" in set_codes2
    assert "Cosmos Elixir" in card_names2
    assert "A-Cosmos Elixir" not in card_names2

    assert "yecl" not in set_codes2
    assert "Thoughtsteal Sprites" not in card_names2


def test_exclude_token_layout(scryfall_data: bundles.ScryfallDataSet) -> None:
    set_codes = {s.code for s in scryfall_data.sets}
    card_names = {c.name for c in scryfall_data.cards}

    assert "p03" in set_codes
    assert "sld" in set_codes
    assert "Goblin" in card_names

    tokens_removed = bundles.filter_cards_and_sets(
        scryfall_data, exclude_card_layouts={ScryCardLayout.TOKEN}
    )
    set_codes2 = {s.code for s in tokens_removed.sets}
    card_names2 = {c.name for c in tokens_removed.cards}
    assert "p03" not in set_codes2
    assert "sld" in set_codes2
    assert "Goblin" not in card_names2


def test_exclude_foreign_only(scryfall_data: bundles.ScryfallDataSet) -> None:
    set_codes = {s.code for s in scryfall_data.sets}
    card_ids = {c.id for c in scryfall_data.cards}

    assert "ren" in set_codes
    assert UUID("81917a2b-9bf6-4aa6-947d-36b0f45d6fe3") in card_ids

    tokens_removed = bundles.filter_cards_and_sets(scryfall_data, exclude_foreing_only=True)
    set_codes2 = {s.code for s in tokens_removed.sets}
    card_ids2 = {c.id for c in tokens_removed.cards}
    assert "ren" not in set_codes2
    assert UUID("81917a2b-9bf6-4aa6-947d-36b0f45d6fe3") not in card_ids2


def test_merge_promos(scryfall_data: bundles.ScryfallDataSet) -> None:
    card_ids_and_sets = {(c.id, c.set) for c in scryfall_data.cards}

    # Promo Tithe Taker
    assert (UUID("848d4dae-1e51-446f-bad0-cdf852970486"), "prna") in card_ids_and_sets
    assert (UUID("4d7969a1-afae-4076-bb1f-fb52014badea"), "prna") in card_ids_and_sets
    # Main set Tithe Taker
    assert (UUID("bd26b7b1-992d-4b8c-bc33-51aab5abdf98"), "rna") in card_ids_and_sets
    # Separate promo set Plains that should not be interleaved
    assert (UUID("8004052e-cb88-4ca6-a563-5396f13f7c6d"), "prw2") in card_ids_and_sets

    promos_merged = bundles.filter_cards_and_sets(scryfall_data, merge_promos=True)
    card_ids_and_sets2 = {(c.id, c.set) for c in promos_merged.cards}

    # Promo Tithe Taker
    assert (UUID("848d4dae-1e51-446f-bad0-cdf852970486"), "rna") in card_ids_and_sets2
    assert (UUID("4d7969a1-afae-4076-bb1f-fb52014badea"), "rna") in card_ids_and_sets2
    # Main set Tithe Taker
    assert (UUID("bd26b7b1-992d-4b8c-bc33-51aab5abdf98"), "rna") in card_ids_and_sets2
    # Separate promo set Plains that should not be interleaved
    assert (UUID("8004052e-cb88-4ca6-a563-5396f13f7c6d"), "prw2") in card_ids_and_sets2


def test_default_exclusions_cover_recent_scryfall_categories(
    scryfall_data: bundles.ScryfallDataSet,
) -> None:
    set_codes = {s.code for s in scryfall_data.sets}
    card_names = {c.name for c in scryfall_data.cards}

    assert {"altc", "macr", "tmc", "tmma", "bro", "znr"} <= set_codes
    assert {
        "Fell Beast's Shriek // Fell Beast's Shriek",
        "Into the Story: Assassin Edition // Story Beats",
        "Ninja Pizza",
        "Elspeth, Knight-Errant Emblem",
        "Steel Seraph",
        "Skyclave Cleric // Skyclave Basilica",
    } <= card_names

    default_filtered = bundles.filter_cards_and_sets(
        scryfall_data,
        exclude_set_types={
            ScrySetType.MEMORABILIA,
            ScrySetType.MINIGAME,
            ScrySetType.TOKEN,
        },
        exclude_card_layouts={
            ScryCardLayout.ART_SERIES,
            ScryCardLayout.DOUBLE_FACED_TOKEN,
            ScryCardLayout.EMBLEM,
            ScryCardLayout.TOKEN,
        },
    )
    set_codes2 = {s.code for s in default_filtered.sets}
    card_names2 = {c.name for c in default_filtered.cards}

    assert "altc" not in set_codes2
    assert "macr" not in set_codes2
    assert "tmma" not in set_codes2
    assert "Fell Beast's Shriek // Fell Beast's Shriek" not in card_names2
    assert "Into the Story: Assassin Edition // Story Beats" not in card_names2
    assert "Elspeth, Knight-Errant Emblem" not in card_names2

    assert {"tmc", "bro", "znr"} <= set_codes2
    assert {
        "Ninja Pizza",
        "Steel Seraph",
        "Skyclave Cleric // Skyclave Basilica",
    } <= card_names2


def test_filtered_set_card_counts_match_remaining_cards(
    scryfall_data: bundles.ScryfallDataSet,
) -> None:
    filtered = bundles.filter_cards_and_sets(
        scryfall_data,
        exclude_set_types={
            ScrySetType.MEMORABILIA,
            ScrySetType.MINIGAME,
            ScrySetType.TOKEN,
        },
        exclude_card_layouts={
            ScryCardLayout.ART_SERIES,
            ScryCardLayout.DOUBLE_FACED_TOKEN,
            ScryCardLayout.EMBLEM,
            ScryCardLayout.TOKEN,
        },
        exclude_digital=True,
        exclude_foreing_only=True,
        merge_promos=True,
    )
    set_code_to_count = Counter(card.set for card in filtered.cards)
    assert set_code_to_count
    assert {set_.code for set_ in filtered.sets} == set(set_code_to_count)
    assert all(set_.card_count == set_code_to_count[set_.code] for set_ in filtered.sets)



"""Data bundle definitions."""

import copy
from collections import Counter
from typing import NamedTuple

from mtg_ssm.scryfall.models import ScryCard, ScryCardLayout, ScryMigration, ScrySet, ScrySetType


class ScryfallDataSet(NamedTuple):
    """Bundle for storing Scryfall data."""

    sets: list[ScrySet]
    cards: list[ScryCard]
    migrations: list[ScryMigration]


def recount_sets(sets: list[ScrySet], cards: list[ScryCard]) -> list[ScrySet]:
    """Return sets with card counts updated to match the provided card list."""
    set_code_to_count = Counter(card.set for card in cards)
    recounted_sets = []
    for set_ in sets:
        card_count = set_code_to_count[set_.code]
        if set_.card_count == card_count:
            recounted_sets.append(set_)
            continue
        recounted_set = copy.copy(set_)
        recounted_set.card_count = card_count
        recounted_sets.append(recounted_set)
    return recounted_sets


def filter_cards_and_sets(  # noqa: C901
    scryfall_data: ScryfallDataSet,
    *,
    exclude_set_types: set[ScrySetType] | None = None,
    exclude_card_layouts: set[ScryCardLayout] | None = None,
    exclude_digital: bool = False,
    exclude_foreing_only: bool = False,
    merge_promos: bool = False,
) -> ScryfallDataSet:
    """Filter a ScryfallDataSet to exclude desired set types, card layouts, and digital only products."""
    accepted_setcodes = set()
    remapped_setcodes = {}
    for set_ in scryfall_data.sets:
        if exclude_set_types and set_.set_type in exclude_set_types:
            continue
        if exclude_digital and set_.digital:
            continue
        if (
            merge_promos
            and set_.set_type == ScrySetType.PROMO
            and set_.parent_set_code is not None
            and set_.code == f"p{set_.parent_set_code}"
        ):
            remapped_setcodes[set_.code] = set_.parent_set_code
            continue
        accepted_setcodes.add(set_.code)

    accepted_cards = []
    nonempty_setcodes = set()
    for card_loop in scryfall_data.cards:
        card = card_loop  # capture loop variable
        while card.set in remapped_setcodes:
            collector_number = card.collector_number
            if collector_number.isdigit():
                collector_number += "p"
            card = copy.copy(card)
            card.set = remapped_setcodes[card.set]
            card.collector_number = collector_number
        if card.set not in accepted_setcodes:
            continue
        if exclude_card_layouts and card.layout in exclude_card_layouts:
            continue
        if exclude_digital and card.digital:
            continue
        if exclude_foreing_only and card.lang != "en":
            continue
        accepted_cards.append(card)
        nonempty_setcodes.add(card.set)

    accepted_sets = [
        s
        for s in scryfall_data.sets
        if s.code in accepted_setcodes and s.code in nonempty_setcodes
    ]
    accepted_sets = recount_sets(accepted_sets, accepted_cards)

    return ScryfallDataSet(
        sets=accepted_sets,
        cards=accepted_cards,
        migrations=scryfall_data.migrations,
    )



#!/usr/bin/env python3
"""Script for managing magic card spreadsheets."""

import argparse
import datetime as dt
from pathlib import Path

import mtg_ssm
import mtg_ssm.serialization.interface as ser_interface
from mtg_ssm.containers import bundles
from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.scryfall import fetcher
from mtg_ssm.scryfall.models import ScryCardLayout, ScrySetType


def epilog() -> str:
    """Generate the argparse help epilog with dialect descriptions."""
    dialect_docs = "available dialects:\n"
    ext_dia_desc = ser_interface.SerializationDialect.dialects()
    for extension, dialect, description in ext_dia_desc:
        dialect_docs += f"  {extension:<8} {dialect:<12} {description}\n"
    return dialect_docs


def set_type_list(value: str) -> set[ScrySetType]:
    """Argparse type to convert a string to a set of Scryfall Set Types."""
    set_types = set()
    for set_str in value.split(","):
        try:
            set_types.add(ScrySetType(set_str))
        except ValueError as err:
            msg = (
                f"{set_str} in {value} is not a valid set_type, please use a comma separated list of values from: "
                + ", ".join(ScrySetType)
            )
            raise argparse.ArgumentTypeError(msg) from err
    return set_types


def card_layout_list(value: str) -> set[ScryCardLayout]:
    """Argparse type to convert a string to a set of Scryfall Card Layouts."""
    card_layouts = set()
    for layout_str in value.split(","):
        try:
            card_layouts.add(ScryCardLayout(layout_str))
        except ValueError as err:
            msg = (
                f"{layout_str} in {value} is not a valid layout_type, please use a comma separated list of values from: "
                + ", ".join(ScryCardLayout)
            )
            raise argparse.ArgumentTypeError(msg) from err
    return card_layouts


def get_args(args: list[str] | None = None) -> argparse.Namespace:
    """Parse and return application arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Magic collection Spreadsheet Manager",
        epilog=epilog(),
    )
    parser.add_argument("--version", action="version", version=mtg_ssm.__version__)

    parser.add_argument(
        "--include-digital",
        default=False,
        action="store_true",
        help="Include digital only sets (e.g. Vintage Masters)",
    )

    parser.add_argument(
        "--include-foreign-only",
        default=False,
        action="store_true",
        help="Include foreign only cards/sets (e.g. Renaissance)",
    )

    parser.add_argument(
        "--separate-promos",
        default=False,
        action="store_true",
        help="Treat set promos as a separate set/tab",
    )

    default_exclude_set_types = {
        ScrySetType.MEMORABILIA,
        ScrySetType.TOKEN,
        ScrySetType.MINIGAME,
    }
    parser.add_argument(
        "--exclude-set-types",
        default=",".join(default_exclude_set_types),
        type=set_type_list,
        help="List of set types to exclude from data as a comma separated list of values from: "
        + ", ".join(ScrySetType),
    )

    default_exclude_card_layouts = {
        ScryCardLayout.ART_SERIES,
        ScryCardLayout.DOUBLE_FACED_TOKEN,
        ScryCardLayout.EMBLEM,
        ScryCardLayout.TOKEN,
    }
    parser.add_argument(
        "--exclude-card-layouts",
        default=",".join(default_exclude_card_layouts),
        type=card_layout_list,
        help="List of card layouts to exclude from data as a comma separated list of values from: "
        + ", ".join(ScryCardLayout),
    )

    parser.add_argument(
        "-d",
        "--dialect",
        nargs=2,
        metavar=("EXTENSION", "DIALECT"),
        action="append",
        default=[],
        help="Mapping of file extensions to serializer dialects. "
        "May be repeated for multiple different extensions.",
    )

    # Commands
    subparsers = parser.add_subparsers(dest="action", title="actions")
    subparsers.required = True

    create = subparsers.add_parser(
        "create", aliases=["c"], help="Create a new, empty collection spreadsheet"
    )
    create.set_defaults(func=create_cmd)
    create.add_argument("collection", type=Path, help="Filename for the new collection")

    update = subparsers.add_parser(
        "update",
        aliases=["u"],
        help="Update cards in a collection spreadsheet, preserving counts",
    )
    update.set_defaults(func=update_cmd)
    update.add_argument("collection", type=Path, help="Filename for the collection to update")

    merge = subparsers.add_parser(
        "merge",
        aliases=["m"],
        help="Merge one or more collection spreadsheets into another. May also be used for format conversions.",
    )
    merge.set_defaults(func=merge_cmd)
    merge.add_argument("collection", type=Path, help="Filename for the target collection")
    merge.add_argument(
        "imports",
        nargs="+",
        type=Path,
        help="Filename(s) for collection(s) to import/merge counts from",
    )

    diff = subparsers.add_parser(
        "diff",
        aliases=["d"],
        help="Create a collection from the differences between two other collections",
    )
    diff.set_defaults(func=diff_cmd)
    diff.add_argument(
        "left",
        type=Path,
        help="Filename for first collection to diff (positive counts)",
    )
    diff.add_argument(
        "right",
        type=Path,
        help="Filename for second collection to diff (negative counts)",
    )
    diff.add_argument("output", type=Path, help="Filename for result collection of diff")

    parsed_args = parser.parse_args(args=args)
    parsed_args.dialect = dict(parsed_args.dialect)
    return parsed_args


def get_oracle(
    *,
    exclude_set_types: set[ScrySetType],
    exclude_card_layouts: set[ScryCardLayout],
    include_digital: bool,
    include_foreign_only: bool,
    separate_promos: bool,
) -> Oracle:
    """Get a card_db with current mtgjson data."""
    scrydata = fetcher.scryfetch()
    scrydata = bundles.filter_cards_and_sets(
        scrydata,
        exclude_set_types=exclude_set_types,
        exclude_card_layouts=exclude_card_layouts,
        exclude_digital=not include_digital,
        exclude_foreing_only=not include_foreign_only,
        merge_promos=not separate_promos,
    )
    return Oracle(scrydata)


def get_serializer(
    dialect_mapping: dict[str, str], path: Path
) -> ser_interface.SerializationDialect:
    """Retrieve a serializer compatible with a given filename."""
    extension = path.suffix.lstrip(".")
    serialization_class = ser_interface.SerializationDialect.by_extension(
        extension, dialect_mapping
    )
    return serialization_class()


def get_backup_path(path: Path) -> Path:
    """Given a filename, return a timestamped backup name for the file."""
    now = dt.datetime.now(dt.timezone.utc).astimezone()
    return path.parent / f"{path.stem}.{now:%Y%m%d_%H%M%S}{path.suffix}"


def get_temp_path(path: Path) -> Path:
    """Given a filename, return a temporary path to use for a new file."""
    return path.parent / f"{path.stem}.tmp{path.suffix}"


def write_file(
    serializer: ser_interface.SerializationDialect,
    collection: MagicCollection,
    path: Path,
) -> None:
    """Write print counts to a file, backing up existing target files."""
    temp_path = get_temp_path(path)
    print(f"Writing to temporary file: {temp_path}")
    serializer.write(temp_path, collection)
    if path.exists():
        backup_path = get_backup_path(path)
        print(f"Backing up existing file to: {backup_path}")
        path.replace(backup_path)
    print(f"Writing collection: {path}")
    temp_path.replace(path)


def create_cmd(args: argparse.Namespace, oracle: Oracle) -> None:
    """Create a new, empty collection."""
    collection = MagicCollection(oracle=oracle, counts={})
    serializer = get_serializer(args.dialect, args.collection)
    write_file(serializer, collection, args.collection)


def update_cmd(args: argparse.Namespace, oracle: Oracle) -> None:
    """Update an existing collection, preserving counts."""
    serializer = get_serializer(args.dialect, args.collection)
    print(f"Reading counts from {args.collection}")
    collection = serializer.read(args.collection, oracle)
    write_file(serializer, collection, args.collection)


def merge_cmd(args: argparse.Namespace, oracle: Oracle) -> None:
    """Merge counts from one or more inputs into a new/existing collection."""
    coll_serializer = get_serializer(args.dialect, args.collection)
    collection = MagicCollection(oracle=oracle, counts={})
    if args.collection.exists():
        print(f"Reading counts from {args.collection}")
        collection = coll_serializer.read(args.collection, oracle)
    for import_path in args.imports:
        input_serializer = get_serializer(args.dialect, import_path)
        print(f"Merging counts from {import_path}")
        collection += input_serializer.read(import_path, oracle)
    write_file(coll_serializer, collection, args.collection)


def diff_cmd(args: argparse.Namespace, oracle: Oracle) -> None:
    """Diff two collections, putting the output in a third."""
    left_serializer = get_serializer(args.dialect, args.left)
    right_serializer = get_serializer(args.dialect, args.right)
    output_serializer = get_serializer(args.dialect, args.output)
    print(f"Diffing counts between {args.left} and {args.right}")
    diff_collection = left_serializer.read(args.left, oracle) - right_serializer.read(
        args.right, oracle
    )
    write_file(output_serializer, diff_collection, args.output)


def main() -> None:
    """Get args and run the appropriate command."""
    args = get_args()
    oracle = get_oracle(
        exclude_set_types=args.exclude_set_types,
        exclude_card_layouts=args.exclude_card_layouts,
        include_digital=args.include_digital,
        include_foreign_only=args.include_foreign_only,
        separate_promos=args.separate_promos,
    )
    args.func(args, oracle)


if __name__ == "__main__":
    main()



"""Container for all data related to a collection."""

from dataclasses import dataclass

from typing_extensions import Self

from mtg_ssm.containers import counts
from mtg_ssm.containers.counts import ScryfallCardCount
from mtg_ssm.containers.indexes import Oracle


@dataclass
class MagicCollection:
    """Collection object for tracking magic cards and counts."""

    oracle: Oracle
    counts: ScryfallCardCount

    def __add__(self, other: "MagicCollection") -> "MagicCollection":
        if not isinstance(other, MagicCollection):
            return NotImplemented
        return MagicCollection(
            oracle=self.oracle,
            counts=counts.merge_card_counts(self.counts, other.counts),
        )

    def __iadd__(self, other: "MagicCollection") -> Self:
        if not isinstance(other, MagicCollection):
            return NotImplemented
        self.counts = counts.merge_card_counts(self.counts, other.counts)
        return self

    def __sub__(self, other: "MagicCollection") -> "MagicCollection":
        if not isinstance(other, MagicCollection):
            return NotImplemented
        return MagicCollection(
            oracle=self.oracle,
            counts=counts.diff_card_counts(self.counts, other.counts),
        )

    def __isub__(self, other: "MagicCollection") -> Self:
        if not isinstance(other, MagicCollection):
            return NotImplemented
        self.counts = counts.diff_card_counts(self.counts, other.counts)
        return self
